from typing import Union, Callable, Sequence
from argparse import ArgumentParser

_complete_bash_script = """
_{name}_comp() {{
    COMPREPLY=()
    if [ ${{COMP_CWORD}} -eq 1 ] ; then
        COMPREPLY=( $(compgen -W "$({cmd} complete)" -- "$2" ) )
    fi
}}
complete -F _{name}_comp  -o bashdefault -o default {cmd}
"""


class Cmd:
    _all: list["Cmd"] = list()

    #: version string
    version: str = ""

    #: ArgumentParser prog
    prog: str = None

    #: generate complete commands if True
    complete: bool = True

    #: complete words
    complete_words: list[str] = ['--help']

    #: Command name
    name: Union[None, str, Sequence[str]] = None

    #: Command Function
    function: Union[str, Callable] = None

    #: Help string (in command list) hidden if None
    help = ""

    #: Help string (in commnd help)
    desc = ""

    #: Complete for cmd
    cmd_complete: list[str] = []

    _parser: ArgumentParser = None

    def __init__(self) -> None:
        super().__init__()
        if isinstance(self.name, str):
            self.name = [self.name]

    def init_parser(self, subparsers):
        params = dict()
        if self.help is not None:
            params['help'] = self.help
        if self.desc:
            params['description'] = self.desc
        parser = subparsers.add_parser(self.name[0], aliases=self.name[1:], **params)
        parser.set_defaults(sub_cmd_func=self.execute_cmd)
        self.add_arguments(parser)

    def add_arguments(self, parser: ArgumentParser):
        """
        Set command arguments

        Exemple::

            parser.add_argument('--value', help="Value")
        """

    def execute_cmd(self, arg):
        if isinstance(self.function, str):
            m, f = self.function.rsplit('.', 1)
            func = getattr(__import__('linux_admin_utils.' + m, globals(), locals(), [f]), f)
        else:
            func = self.function

        func(**vars(arg))

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls.name is not None:
            cls._all.append(cls())

    @classmethod
    def complete_init(cls, subparsers):
        parser = subparsers.add_parser("complete")
        parser.set_defaults(sub_cmd_func=cls.complete_cmd)
        parser = subparsers.add_parser("complete-bash")
        parser.set_defaults(sub_cmd_func=cls.complete_bash)

    @classmethod
    def complete_cmd(cls, _):
        words = set(cls.complete_words)
        if cls.version:
            words.add('--version')
        for cmd in cls._all:
            words.add(cmd.name[0])
            for w in cmd.cmd_complete:
                words.add(w)

        print(' '.join(sorted(words)))

    @classmethod
    def complete_bash(cls, _):

        print(_complete_bash_script.format(
            cmd=cls._parser.prog,
            name=cls._parser.prog.replace("-", "_").replace("/", "_")
        ))

    @classmethod
    def main(cls):
        cls._parser = ArgumentParser(prog=cls.prog)
        if cls.version:
            cls._parser.add_argument('--version', "-V", action='version', version=cls.version)
        cls._parser.set_defaults(sub_cmd_func=None)

        cls.main_parser(cls._parser)

        subparsers = cls._parser.add_subparsers(metavar="", help='', title="Commands")
        for i in Cmd._all:
            i.init_parser(subparsers)

        if cls.complete:
            cls.complete_init(subparsers)

        arg = cls._parser.parse_args()

        cls.main_args(arg)

        if not arg.sub_cmd_func:
            cls._parser.print_help()
            return
        arg.sub_cmd_func(arg)

    @classmethod
    def main_parser(cls, parser: ArgumentParser):
        pass

    @classmethod
    def main_args(cls, args):
        pass



from setuptools import setup

setup(
    name='xbutils',
    version='0.1.1',
    packages=['xbutils'],
    url='https://github.com/exbee49/xbutils',
    license='WTFPL',
    author='xb49',
    author_email='',
    description='Python utilities',
    long_description="Some Python utility classes",
    long_description_content_type="text/plain",

    classifiers=[
        "Development Status :: 4 - Beta",
        # "license :: Public Domain :: WTFPLv2",
        "Programming Language :: Python :: 3 :: Only"
    ]
)



"""
cfg module
"""
from pathlib import Path
from typing import Union


class SimpleConfig:
    """Simple Config File.

    The attributes must be defined with type::

        class Cfg(SimpleConfig):
            int_value: int = 1
            bool_value: bool = True
            str_value: str = "toto"
            path_value: Path = Path("/usr")
            float_value: float = 1.


    """

    @staticmethod
    def _error(text):
        """ override to print error messages"""
        print("*ERR*", text)

    def _init(self):
        """ override to init value (called in __init__)"""
        pass

    def _before_read(self):
        """ hook called before read"""
        pass

    def _after_read(self):
        """ hook called after read"""
        pass

    def __init__(self) -> None:
        super().__init__()
        self._init()

    def read_config(self, cfg_file: Union[Path, str], must_exist=False) -> bool:
        self._before_read()
        cfg_file = Path(cfg_file).expanduser().resolve()
        if not cfg_file.exists():
            if not must_exist:
                return True
            self._error(f"Missing config file {cfg_file}")
            return False
        ok = True
        local_dict = dict()
        try:
            exec(cfg_file.read_text(encoding="utf8"), dict(Path=Path, config_directory=cfg_file.resolve().parent),
                 local_dict)
        except Exception as v:
            self._error(f"Reading Config: {v}")
            ok = False

        valid_attrs = dict(
            (k, v) for k, v in self.__annotations__.items()
            if k[:1] != '_' and v in (str, int, bool, float, Path)
        )

        for k, v in local_dict.items():
            if k[:1] == "_":
                continue
            if k not in valid_attrs:
                print(f"*WRN* Unknown config value {k}")
                ok = False
                continue
            # noinspection PyTypeHints
            if valid_attrs[k] == Path:
                if isinstance(v, str):
                    v = Path(v)
                    if '~' in str(v):
                        v = v.expanduser()
                elif not isinstance(v, Path):
                    self._error(f"Bad type for config value {k} must be str or Path")
                    ok = False
                    continue
            elif valid_attrs[k] == float:
                if isinstance(v, int):
                    v = float(v)
                elif not isinstance(v, float):
                    self._error(f"Bad type for config value {k} must be a float")
                    ok = False
                    continue
            elif not isinstance(v, valid_attrs[k]):
                self._error(f"*WRN* Bad type for config value {k} must be {valid_attrs[k]}")
                ok = False
                continue
            setattr(self, k, v)
        self._after_read()
        return ok
